'use strict';

const base = require('unique-model').backend.base;
const mysql2 = require('mysql2/promise');
const { SqlBuilder } = require('./sql/builder');

class Dao extends base.Dao {
  constructor(session, sqlBuilder, connection, modelClass) {
    super(session);

    this._sqlBuilder = sqlBuilder;
    this._connection = connection;
    this.schema = modelClass._meta.schema;
    this.modelClass = modelClass;
  }

  getObjectIdKey() {
    return 'id';
  }

  // interface
  toObjectId(idString) {
    return Number(idString);
  }

  fromObjectId(idString) {
    return Number(idString);
  }

  // override
  async createAsync(backendModel) {
    const statement = this._sqlBuilder.generateInsertStatement(backendModel);
    const [insertResult] = await this._connection.execute(statement.sql, statement.values);

    backendModel.id = insertResult.insertId;

    return backendModel;
  }

  // override
  async insertManyAsync(plainObjects) {
    const insertResult = await this._collection.insertMany(plainObjects);

    return insertResult.ops;
  }

  countAsync(condition) {
    return this._collection.count(condition);
  }

  // override
  distinctAsync(field, condition) {
    return this._collection.distinct(field, condition);
  }

  aggregateAsync(condition) {
    return this._collection.count(condition);
  }

  // override
  queryAsync(condition) {
    const cursor = this._collection.find(condition);

    return new Query(this, cursor);
  }

  // override
  findOneAsync(condition, options) {
    return this._collection.findOne(condition, options);
  }

  // override
  updateAsync(condition, doc, options) {
    return this._collection.updateMany(condition, doc, options);
  }

  // override
  updateOneAsync(condition, doc, options) {
    return this._collection.updateOne(condition, doc, options);
  }

  // override
  async findOneAndUpdateAsync(condition, doc, options) {
    const result = await this._collection.findOneAndUpdate(condition, doc, options);
    if ( !result.ok ) {
      throw new Error(result.lastErrorObject);
    }

    return result.value;
  }

  // override
  removeAsync(condition) {
    const statement = this._sqlBuilder.generateDeleteStatement(condition);

    return this._connection.execute(statement.sql, statement.values);
  }
}

class Session extends base.Session {
  constructor(options) {
    super(options);

    this._options = options.options;
    this._connection = null;
  }

  async connect() {
    this._connection = await mysql2.createPool(this._options);
    await this.checkConnection(this._connection);
  }

  async checkConnection(connection) {
    const [rows] = await connection.execute('SELECT 1+1 AS result');

    if ( rows.length !== 1 || !rows[0] || rows[0].result !== 2 ) {
      throw new Error('Query error');
    }
  }

  // override
  createDao(modelClass) {
    const tableName = modelClass._meta.tableName;
    const sqlBuilder = new SqlBuilder(tableName);

    return new Dao(this, sqlBuilder, this._connection, modelClass);
  }

  // override
  close() {
    if ( !this._connection ) {
      return false;
    }

    this._connection.end();

    return true;
  }
}

module.exports = {
  Session
};
