'use strict';

const { generateWhereExpression } = require('./expr/where');
const { generateSetExpressions } = require('./expr/set');
const { ExpressionType } = require('./expr/common');

class SqlBuilder {
  constructor(tableName) {
    this._tableName = tableName;
    this._skip = null;
    this._limit = null;
    this._sort = null;
  }

  skip(skip) {
    this._skip = skip;
  }

  limit(limit) {
    this._limit = limit;
  }

  sort(sort) {
    this._sort = sort;
  }

  generateSelectStatement(attributes, condition) {
    const attributesClause = this.generateAttributesClause(attributes);
    const fromClause = this.generateFromClause();
    const whereClause = this.generateWhereClause(condition);
    const orderClause = this.generateOrderClause();
    const limitClause = this.generateLimitClause();

    return this.combineClauses('SELECT', [
      attributesClause,
      fromClause,
      whereClause,
      orderClause,
      limitClause
    ]);
  }

  generateInsertStatement(object) {
    const fields = Object.keys(object);
    const values = fields.map(field => {
      return object[field];
    });
    const placeholders = fields.map(() => {
      return '?';
    });

    const fieldsExpression = fields.join(', ');
    const placeholderExpression = placeholders.join(', ');
    const sql = `INSERT INTO ${this._tableName} (${fieldsExpression}) VALUES (${placeholderExpression})`;

    return {
      sql,
      values
    };
  }

  generateUpdateStatement(condition, updater) {
    const tableClause = {
      sql: this._tableName
    };
    const setClause = this.generateSetClause(updater);
    const whereClause = this.generateWhereClause(condition);

    return this.combineClauses('UPDATE', [
      tableClause,
      setClause,
      whereClause
    ]);
  }

  generateDeleteStatement(condition) {
    const fromClause = this.generateFromClause();
    const whereClause = this.generateWhereClause(condition);

    return this.combineClauses('DELETE', [
      fromClause,
      whereClause
    ]);
  }

  combineClauses(prefix, clauses) {
    const sqlParts = [];
    let values = [];

    for ( const clause of clauses ) {
      if ( clause.keyword ) {
        sqlParts.push(clause.keyword);
      }

      if ( clause.sql ) {
        sqlParts.push(clause.sql);
      }

      if ( clause.keywords ) {
        const keywordsCount = clause.keywords.length;
        const sqlsCount = clause.sqls.length;

        if ( keywordsCount !== sqlsCount ) {
          throw new Error('The keywords count must equal to sqls count');
        }

        for ( let keywordIndex = 0; keywordIndex < keywordsCount; ++ keywordIndex ) {
          sqlParts.push(clause.keywords[keywordIndex]);
          sqlParts.push(clause.sqls[keywordIndex]);
        }
      }

      if ( clause.values ) {
        values = values.concat(clause.values);
      }
    }

    const sql = `${prefix} ${sqlParts.join(' ')}`;

    return {
      sql,
      values
    };
  }

  generateAttributesClause(attributes) {
    const completeAttributes = attributes.map(attribute => {
      return `${this._tableName}.${attribute}`;
    });

    return {
      sql: completeAttributes.join(', ')
    };
  }

  generateFromClause() {
    return {
      keyword: 'FROM',
      sql: this._tableName
    };
  }

  generateSetClause(updater) {
    return generateSetExpressions(updater);
  }

  generateWhereClause(condition) {
    const expressions = generateWhereExpression(condition);
    const statementParts = [];
    let statementValues = [];
    for ( const expression of expressions ) {
      if ( expression.type === ExpressionType.Operator ) {
        statementParts.push(expression.key);
      }
      else if ( expression.type === ExpressionType.Expression ) {
        statementParts.push(`${expression.key} ${expression.op} ${expression.placeholder}`);
        if ( 'value' in expression ) {
          const sqlValue = expression.value;

          if ( sqlValue instanceof Array ) {
            statementValues = statementValues.concat(sqlValue);
          }
          else {
            statementValues.push(sqlValue);
          }
        }
      }
    }

    const sql = statementParts.join(' ');
    if ( sql === '( )') {
      return {};
    }

    return {
      keyword: 'WHERE',
      sql: sql,
      values: statementValues
    };
  }

  generateOrderClause() {
    if ( !this._sort ) {
      return {
      };
    }

    const sortKeys = Object.keys(this._sort);
    const sortParts = [];
    for ( const sortKey of sortKeys ) {
      const sortValue = this._sort[sortKey];

      if ( sortValue === 1 ) {
        sortParts.push(`ASC ${sortKey}`);
      }
      else if ( sortValue === -1 ) {
        sortParts.push(`DESC ${sortKey}`);
      }
      else {
        throw new Error('The sort field value must be 1/-1');
      }
    }

    return {
      keyword: 'ORDER BY',
      sql: sortParts.join(', ')
    };
  }

  generateLimitClause() {
    const keywords = [];
    const sqls = [];

    if ( typeof this._limit === 'number' ) {
      keywords.push('LIMIT');
      sqls.push(String(this._limit));
    }

    if ( typeof this._skip === 'number' ) {
      keywords.push('OFFSET');
      sqls.push(String(this._skip));
    }

    if ( keywords.length === 0 ) {
      return {
      };
    }

    return {
      keywords,
      sqls
    };
  }
}

module.exports = {
  SqlBuilder
};
