'use strict';

const { ComparisonOperators } = require('../op/comparison');
const { LogicalOperators } = require('../op/logical');
const regexGenerator = require('../op/regex');
const { ExpressionType } = require('./common');

const ObjectPrototype = Object.getPrototypeOf({});

function generateWhereExpression(condition) {
  return generateSubExpressions(generateObjectExpressions.bind(null, condition, 'AND'));
}

function generateSubExpressions(generator) {
  let expressions = [{
    type: ExpressionType.Operator,
    key: '('
  }];

  expressions = generator(expressions);

  if ( expressions.length > 1 ) {
    expressions.pop();
  }

  if ( expressions.length === 2 ) {
    expressions.shift();
  }
  else {
    expressions.push({
      type: ExpressionType.Operator,
      key: ')'
    });
  }

  return expressions;
}

function generateObjectExpressions(condition, concatOperator) {
  return generateSubExpressions(
    generateObjectExpressionsImpl.bind(null, condition, concatOperator));
}

const ExpressionGenerators = [{
  key(key) {
    return !key.startsWith('$');
  },
  generator: generatePropertyExpressions
}, {
  key(key) {
    return LogicalOperators.has(key);
  },
  generator: generateLogicalExpressions
}];

function generateObjectExpressionsImpl(condition, concatOperator, expressions) {
  const keys = Object.keys(condition);

  for ( const key of keys ) {
    let keyProcessed = false;
    for ( const expressionGenerator of ExpressionGenerators ) {
      if ( expressionGenerator.key(key) ) {
        expressions = expressionGenerator.generator(condition, key, concatOperator, expressions);

        keyProcessed = true;
        break;
      }
    }

    if ( !keyProcessed ) {
      throw new Error(`The SQL Builder can not support operator ${key} now.`);
    }
  }

  return expressions;
}

function generateArrayExpressions(condition, concatOperator, logicalOperator) {
  return generateSubExpressions(
    generateArrayExpressionsImpl.bind(null, condition, concatOperator, logicalOperator));
}

function generateArrayExpressionsImpl(conditionArray, concatOperator, logicalOperator, expressions) {
  if ( !(conditionArray instanceof Array) ) {
    throw Error(`The sub condition of ${logicalOperator.key} must be an array`);
  }

  expressions = [{
    type: ExpressionType.Operator,
    key: '('
  }];

  for ( const condition of conditionArray ) {
    expressions = expressions.concat(
      generateObjectExpressions(condition, concatOperator));
    expressions.push({
      type: ExpressionType.Operator,
      key: logicalOperator.concat
    });
  }

  return expressions;
}

function generateLogicalExpressions(condition, key, concatOperator, expressions) {
  // throw new Error(`The SQL Builder can not support logical operators now.`);
  const logicalOperator = LogicalOperators.get(key);
  const subCondition = condition[key];

  let subExpressions = [];
  if ( logicalOperator.withArray ) {
    subExpressions = generateArrayExpressions(subCondition, concatOperator, logicalOperator);
  }
  else {
    subExpressions = generateObjectExpressions(subCondition, concatOperator);
  }

  if ( logicalOperator.prefix ) {
    expressions.push({
      type: ExpressionType.Operator,
      key: logicalOperator.prefix
    });
  }

  expressions = expressions.concat(subExpressions);
  expressions.push({
    type: ExpressionType.Operator,
    key: concatOperator
  });

  return expressions;
}

function generatePropertyExpressions(condition, key, concatOperator, expressions) {
  const value = condition[key];
  let sqlOp = '=';
  let sqlPlaceholder = '?';
  let sqlValue = value;
  let noValue = false;

  if ( value === null ) {
    sqlValue = 'NULL';
  }
  else if ( value === undefined ) {
    sqlValue = 'NULL';
  }
  else if ( Object.getPrototypeOf(value) === ObjectPrototype ) {
    if ( regexGenerator.isRegexCondition(value) ) {
      return regexGenerator.generateRegexExpressions(value, key, concatOperator, expressions);
    }

    [ sqlOp, sqlValue, sqlPlaceholder, expressions, noValue ] =
      generateComparisonExpression(expressions, key, value, sqlOp, sqlValue, sqlPlaceholder, concatOperator);

    if ( noValue ) {
      return expressions;
    }
  }

  expressions.push({
    type: ExpressionType.Expression,
    key: key,
    op: sqlOp,
    placeholder: sqlPlaceholder,
    value: sqlValue
  });

  expressions.push({
    type: ExpressionType.Operator,
    key: concatOperator
  });

  return expressions;
}

function generateComparisonExpression(expressions, key, value, sqlOp, sqlValue, sqlPlaceholder, concatOperator) {
  let noValue = false;

  const operatorKey = parseComparisonOperator(value);
  if ( !ComparisonOperators.has(operatorKey) ) {
    throw new Error(`Operator ${key} is not a comparison operator`);
  }

  sqlOp = ComparisonOperators.get(operatorKey);
  sqlValue = value[operatorKey];

  if ( operatorKey === '$in' || operatorKey === '$nin' ) {
    sqlPlaceholder = generateArrayPlaceholder(sqlValue);
    sqlPlaceholder = `(${sqlPlaceholder})`;

    if ( sqlPlaceholder === '()' ) {
      expressions.push({
        type: ExpressionType.Expression,
        key: key,
        op: sqlOp,
        placeholder: sqlPlaceholder
      });

      expressions.push({
        type: ExpressionType.Operator,
        key: concatOperator
      });

      noValue = true;
    }
  }

  return [
    sqlOp,
    sqlValue,
    sqlPlaceholder,
    expressions,
    noValue
  ];
}

function parseComparisonOperator(obj) {
  const keys = Object.keys(obj);

  if ( keys.length !== 1 ) {
    throw new Error('The comparison object must have one key/value');
  }

  return keys[0];
}

function generateArrayPlaceholder(array) {
  if ( !(array instanceof Array) ) {
    throw new Error('The value must be an array');
  }

  const arrayLength = array.length;

  let placeholder = '?, '.repeat(arrayLength);
  if ( placeholder.length > 0 ) {
    placeholder = placeholder.slice(0, -2);
  }

  return placeholder;
}

module.exports = {
  generateWhereExpression
};
